/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: colors.c,v 1.7 2000/08/27 18:32:51 malekith Exp $ */

#include "bot.h"
#include "var.h"

static int color(int fg, int bg)
{
	if (fg == bg)
		return COLOR(LIGHTGRAY, BLACK);
	else
		return COLOR(fg, bg);
}

static int parse_ansi(char *p)
{
	int fg = LIGHTGRAY, bg = BLACK;
	while (*p) {
		switch (atoi(p)) {
		case 21:
		case 22:
		case 27:
		case 24:
		case 0:
			fg = LIGHTGRAY;
			bg = BLACK;
			break;
		case 1:
			fg |= 8;
			break;
		case 2:
			fg = DARKGRAY;
			break;
		case 4:
			fg = CYAN;
			break;
		case 7:
			fg = BLACK;
			bg = WHITE;
			break;
		case 30:
			fg = BLACK;
			break;
		case 31:
			fg = RED;
			break;
		case 32:
			fg = GREEN;
			break;
		case 33:
			fg = BROWN;
			break;
		case 34:
			fg = BLUE;
			break;
		case 35:
			fg = MAGENTA;
			break;
		case 36:
			fg = CYAN;
			break;
		case 37:
			fg = LIGHTGRAY;
			break;
		case 40:
			bg = BLACK;
			break;
		case 41:
			bg = RED;
			break;
		case 42:
			bg = GREEN;
			break;
		case 43:
			bg = BROWN;
			break;
		case 44:
			bg = BLUE;
			break;
		case 45:
			bg = MAGENTA;
			break;
		case 46:
			bg = CYAN;
			break;
		case 47:
			bg = WHITE;
			break;
		case 49:
			bg = BLACK;
			break;
		}
		while (isdigit(*p))
			p++;
		if (*p != ';')
			break;
		p++;
	}
	return color(fg,bg);
}

int map_color(char *p, int n, int def)
{
	struct color {
		char *name;
		int c;
	} colors[] = {
		{ "BLACK", BLACK }, 
		{ "BLUE", BLUE }, 
		{ "GREEN", GREEN }, 
		{ "CYAN", CYAN }, 
		{ "RED", RED }, 
		{ "MAGENTA", MAGENTA },
		{ "BROWN", BROWN }, 
		{ "LIGHTGRAY", LIGHTGRAY }, 
		{ "DARKGRAY", DARKGRAY }, 
		{ "LIGHTBLUE", LIGHTBLUE }, 
		{ "LIGHTGREEN", LIGHTGREEN }, 
		{ "LIGHTCYAN", LIGHTCYAN }, 
		{ "LIGHTRED", LIGHTRED }, 
		{ "LIGHTMAGENTA", LIGHTMAGENTA }, 
		{ "YELLOW", YELLOW }, 
		{ "WHITE", WHITE },
		
		{ "lgray", LIGHTGRAY },
		{ "gray", DARKGRAY },
		{ "lblue", LIGHTBLUE },
		{ "lcyan", LIGHTCYAN },
		{ "lred", LIGHTRED },
		{ "lmagenta", LIGHTMAGENTA },
		{ "norm", LIGHTGRAY },
		{ 0, 0 }
	};
	int i;
	for (i = 0; colors[i].name; i++)
		if (strncasecmp(colors[i].name, p, n) == 0)
			return colors[i].c;
	return def;
}

static int parse_literal(char *p)
{
	char *e;
	int fg = LIGHTGRAY, bg = BLACK;

	if (*p == '{')
		p++;
	for (e = p; *e && *e != '}' && *e != '\n' && *e != ':'; e++)
		/*nothing*/ ;
	fg = map_color(p, e - p, fg);
	if (*e == ':') {
		p = ++e;
		for (e = p; *e && *e != '}' && *e != '\n'; e++)
			/*nothing*/ ;
		bg = map_color(p, e - p, bg);
	}
	return color(fg, bg);
}

static int parse_mirc(char *p)
{
	int n;
	static int mirc_colors[] = {
		LIGHTGRAY, DARKGRAY, BLUE, GREEN, RED, LIGHTRED, MAGENTA,
		BROWN, YELLOW, GREEN, CYAN, LIGHTCYAN, LIGHTBLUE,
		LIGHTMAGENTA, DARKGRAY, WHITE, WHITE
	};
	n = atoi(p);
	if (n < 0)
		n = 0;
	if (n > 15)
		n = 0;
	return color(mirc_colors[n], BLACK);
}

static char *xline_do_compile(char *p, char *d, int lev);
static int xline_count_len(char *p, int lev);

static int compute_thame_entry(char **pp, char *target, int lev)
{
	char *p, *e;
	int len;

	if (lev > 16)
		return 0;
		
	p = *pp;
	while (*p && *p != '\n' && *p != '>')
		p++;
	e = xstrndup(*pp, p - *pp);
	if (*p == '>')
		p++;
	*pp = p;
	p = hsv("thame", e)->var_string;
	xfree(e);
	len = xline_count_len(p, ++lev);

	if (target)
		xline_do_compile(p, target, lev);
		
	return len;
}

static char *xline_do_compile(char *p, char *d, int lev)
{
	int xbold=0, xunder=0, xrev=0,xnorm=0;
	
#define ADD(a,b,v)		\
	p++;			\
	*d++ = '\e'; 		\
	if ((x ## v = !x ## v)) 	\
		*d++ = COLOR(a,b); 	\
	else			\
		*d++ = COLOR(LIGHTGRAY, BLACK);	\
	break;
	
	while (*p) {
		switch (*p) {
		case 2: ADD(WHITE,BLACK,bold);
		case 31: ADD(CYAN,BLACK,under);
		case 22: ADD(BLACK,WHITE,rev);
		case 15: ADD(LIGHTGRAY,BLACK,norm);
		case 3:
			if (isdigit(p[1])) {
				*d++ = '\e';
				*d++ = parse_mirc(++p);
				if (isdigit(*++p))
					p++;
			} else 
				*d++ = *p++;
			break;
		case '\e':
			switch (*++p) {
			case '[':
				*d++ = '\e';
				*d++ = parse_ansi(p);
				while (isdigit(*p) || *p == ';')
					p++;
				if (*p == 'm')
					p++;
				break;
			case '{':
				*d++ = '\e';
				*d++ = parse_literal(p);
				while (*p && *p != '\n' && *p != '}')
					p++;
				if (*p == '}')
					p++;
				break;
			case '<':
				p++;
				d += compute_thame_entry(&p, d, lev);
				break;
			default:
				strcpy(d,"\e\x70" "ESC\e\x07");
				d = strchr(d,0);
				break;
			}
			break;
		default:
			*d++ = *p++;
			break;
		case '\n':
			p++;
			goto stop;
		}
	}
stop:
	*d = 0;
	return p;
}

static int xline_count_len(char *p, int lev)
{
	int len;
	
	len = 0;
	while (*p) {
		switch (*p) {
		case 3:
			if (isdigit(p[1])) {
				len += 2;
				p += 2;
				if (isdigit(*p))
					p++;
			} else {
				len++;
				p++;
			}
			break;
		case 2:
		case 31:
		case 22:
		case 15:
			p++;
			len += 2;
			break;
		case '\e':
			len += 2;
			switch (*++p) {
			case '[':
				while (isdigit(*p) || *p == ';')
					p++;
				if (*p == 'm')
					p++;
				break;
			case '{':
				while (*p && *p != '\n' && *p != '}')
					p++;
				if (*p == '}')
					p++;
				break;
			case '<':
				len -= 2;
				p++;
				len += compute_thame_entry(&p, 0, lev);
				break;
			default:
				len += 5;
				break;
			}
			break;
		default:
			len++;
			p++;
			break;
		case '\n':
			goto stop;
		}
	}
stop:
	return len;
}

/* this starts from lev==2 to avoid any possibility of buffer overflow */
char *line_do_compile(char *p, char *d)
{
	return xline_do_compile(p, d, 2);
}

int line_count_len(char *p)
{
	return xline_count_len(p, 0);
}
