/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: irc.c,v 1.8 2000/08/27 18:32:51 malekith Exp $ */

#include "bot.h"
#include "proc.h"
#include "var.h"
#include "script_ext.h"
#include "proto.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>

#ifdef IPv6
static int use6;
static struct sockaddr_in6 my_addr6;
#endif
static struct sockaddr_in my_addr;

struct server {
	int fd;
	char *name;
	int port;
	char *buffer;
	int buflen, got;
};

static struct server global_server;

struct irc_line {
	char *server;
	int numeric;
	char *nick, *user, *host;
	char *command;
	int argc;
	char **argv;		/* argv[0] == command */
};

#ifdef BOGUS_HSTRERROR
char *hstrerror(int x)
{
	(void)x;
	return "Resolver error";
}
#endif

static void xperror(const char *s)
{
	prtf(0, "\e<error> system error: %s: %s", s, strerror(errno));
}

#ifdef IPv6
static int resolv6(const char *host, int port, struct sockaddr_in6 *sin)
{
	struct hostent *e;
	char *addr;
	
	prtf(0, "\e<info> resolving hostname (IPv6) %s", host);
	
	e = gethostbyname(host);
	if (e == 0) {
		prtf(0, "\e<error> resolv error [home] %s: %s",
		      host, hstrerror(h_errno));
		return -1;
	}
	if (e->h_addrtype == AF_INET) {
		/* rfc2373 #2.5.4 */
		memset(&sin->sin6_addr, 0, sizeof(sin->sin6_addr));
		sin->sin6_addr.s6_addr16[5] = 0xffff;
		memcpy(&sin->sin6_addr.s6_addr32[3], e->h_addr_list[0], 4);
	} else
		memcpy(&sin->sin6_addr, e->h_addr_list[0],
		       sizeof(sin->sin6_addr));
		       
	sin->sin6_port = htons(port);
	sin->sin6_family = AF_INET6;
	
	addr = show_addr((struct sockaddr*)sin, sizeof(*sin));
	prtf(0, "\e<info> host %s resolved as %s", host, addr);
	xfree(addr);
	
	return 0;
}
#endif

static int resolv(const char *host, int port, struct sockaddr_in *sin)
{
	struct hostent *e;
	char *addr;

	prtf(0, "\e<info> resolving hostname (IPv4) %s", host);
	
	e = gethostbyname(host);
	if (e == 0) {
		prtf(0, "\e<error> resolv error [home] %s: %s",
		      host, hstrerror(h_errno));
		return -1;
	}
	
	memcpy(&sin->sin_addr, e->h_addr_list[0], 4);
	
	sin->sin_port = htons(port);
	sin->sin_family = AF_INET;
	
	addr = show_addr((struct sockaddr*)sin, sizeof(*sin));
	prtf(0, "\e<info> host %s resolved as %s", host, addr);
	xfree(addr);
	
	return 0;
}

static int make_socket(const char *home)
{
	int sock;
	struct sockaddr_in sin;
#ifdef IPv6
	struct sockaddr_in6 sin6;

	sock = socket(AF_INET6, SOCK_STREAM, 0);
	
	/* if failed, fallback to v4 */
	use6 = sock != -1;
	if (sock == -1)
#endif
		sock = socket(AF_INET, SOCK_STREAM, 0);
		
	if (sock == -1) {
		xperror("socket");
		return -1;
	}

	if (home == 0 || *home == 0)
		return sock;

#ifdef IPv6
	if (use6) {
		if (resolv6(home, 0, &sin6))
			goto shit;
		
		if (bind(sock, (struct sockaddr*)&sin6, sizeof(sin6))) { 
			xperror("bind");
			goto shit;
		}
	} else
#endif
	{
		if (resolv(home, 0, &sin))
			goto shit;
		if (bind(sock, (struct sockaddr*)&sin, sizeof(sin))) { 
			xperror("bind");
			goto shit;
		}
	}

	return sock;

shit:
	close(sock);
	return -1;
}

static int connect_socket(int sock, const char *host, int port)
{
	struct sockaddr_in sin;
#ifdef IPv6
	struct sockaddr_in6 sin6;
#endif

	if (sock == -1)
		return -1;
		
#ifdef IPv6
	if (use6) {
		if (resolv6(host, port, &sin6))
			goto shit;
		prt(0, "\e<info> connecting v6...");
		if (connect(sock, (struct sockaddr*)&sin6, sizeof(sin6))) {
			xperror("connect");
			goto shit;
		}	
	} else
#endif
	{
		if (resolv(host, port, &sin))
			goto shit;
		prt(0, "\e<info> connecting...");
		if (connect(sock, (struct sockaddr*)&sin, sizeof(sin))) {
			xperror("connect");
			goto shit;
		}
	}

	return sock;

shit:
	close(sock);
	return -1;
}

static void fill_my_addr(int sock)
{
	unsigned int size;
	struct sockaddr *addr;
	
	if (sock == -1)
		return;
		
#ifdef IPv6
	if (use6) {
		size = sizeof(my_addr6);
		addr = (struct sockaddr*)&my_addr6;
	} else
#endif
	{
		size = sizeof(my_addr);
		addr = (struct sockaddr*)&my_addr;
	}
	
	if (getsockname(sock, addr, &size))
		xperror("getsockname");
}


static void kill_server(struct server *s, int do_close)
{
	if (s->fd == -1)
		return;
		
	prtf(0, "\e<info> disconnecting from %s : %d ...",
		s->name, s->port);
		
	if (do_close)
		xclose(s->fd);
	
	xfree(s->name);
	s->name = 0;
	
	s->fd = -1;
}


static char *get_irc_arg(char **pp)
{
	char *p = *pp, *start;

	if (p == 0)
		return 0;
		
	while (*p == ' ')
		p++;
		
	if (*p == 0)
		return 0;
		
	if (*p == ':') {
		*pp = 0; 
		return ++p;
	}
	
	start = p;
	
	while (*p && *p != ' ')
		p++;
	*pp = *p ? p : 0;
	return start;
}

static void split_args(struct irc_line *l, char *p)
{
	char *start;
	char **o;
	
	start = p;
	
	l->argc = 1;
	while (get_irc_arg(&p))
		l->argc++;
		
	o = l->argv = (char**)xmalloc((l->argc + 1) * sizeof(char*));
	
	*o++ = l->command;
	p = start;
	
	while ((*o++ = get_irc_arg(&p)))
		if (p)
			*p++ = 0;
}

static void fill_user_host(struct irc_line *l)
{
	char *s;

	s = l->server;
	if (s == 0)
		return;
		
	if ((l->user = strchr(s, '!'))) {
		l->nick = l->server;
		l->server = 0;
		*l->user++ = 0;
		s = l->user;
	} else
		l->nick = "";
		
	if ((l->host = strchr(s, '@'))) {
		*l->host++ = 0;
		if (l->server) {
			l->nick = l->server;
			l->server = 0;
		}
	} else
		l->host = "";

	/* if server name doesn't contain dots lets guess it's 
	 * a nickname, most servers will send complete
	 * nick!user@host anyway ... */
	if (l->server && strchr(l->server, '.') == 0) {
		l->nick = l->server;
		l->server = 0;
	}
}

static void process_line(struct server *s, struct irc_line *l)
{
	int i;
	char **o;
	
	fill_user_host(l);

	set_sv(hv("input", "server_name"), s->name);
	set_iv(hv("input", "server_port"), s->port);
	set_iv(hv("input", "argc"), l->argc);
	set_iv(hv("input", "numeric"), l->numeric);
	set_sv(hv("input", "command"), l->command);
	set_sv(hv("input", "server"), l->server);
	set_sv(hv("input", "nick"), l->nick);
	set_sv(hv("input", "user"), l->user);
	set_sv(hv("input", "host"), l->host);

	o = l->argv;
	for (i = 0; i < 20; i++) {
		char *s;
		s = xf("argv%d", i);
		set_sv(hv("input", s), o && *o ? *o++ : "");
		xfree(s);
	}

	eval("got_input();");
}

static char *split_input(struct server *s, int force_all)
{
	char *p, *e;
	struct irc_line line;

	memset(&line, 0, sizeof(line));
	
	for (p = s->buffer; p && *p; p = e) {
		e = strchr(p, '\n');
		if (e) {
			if (*--e == '\r')
				*e++ = 0;
			else
				e++;
			*e++ = 0;
		} else {
			e = strchr(p, '\r');
			if (e)
				*e++ = 0;
			else if (!force_all)
				break;
		}
			
		while (*p == ' ')
			p++;
		/* check for prefix, and treat is as server for now */
		if (*p == ':') {
			p++;
			
			line.server = p;
			while (*p && *p != ' ')
				p++;
			if (*p)
				*p++ = 0;
			while (*p == ' ')
				p++;
		}

		/* now find out what the command is */

		/* other forms of numeric are rfc illegal */
		if (p[0] >= '0' && p[0] <= '9'
		    && p[1] >= '0' && p[1] <= '9'
		    && p[2] >= '0' && p[2] <= '9'
		    && p[3] == ' ')
		    	line.numeric = atoi(p);
		 
		line.command = p;
		
		/* according to rfc we should only allow 
		 * letters ... */
		while (*p && *p != ' ')
			p++;
		if (*p) {
			*p++ = 0;
			split_args(&line, p);
		}

		/* and now do something with parsed line ... */
		process_line(s, &line);
	}

	if (p && *p)
		return p;
	else
		return 0;
}

#define START_LEN 4096
#define CHUNK_LEN 1024
static int server_handler(int fd, void *data)
{
	struct server *s;
	int n, eof = 0;
	char *remaining;
	
	s = (struct server*)data;
	
	if (s->buffer == 0)
		s->buffer = (char*)xmalloc((s->buflen = START_LEN) + 1);

	n = read(fd, s->buffer + s->got, s->buflen - s->got);
	
	if (n == s->buflen) {
		do {
			s->buffer = xrealloc(s->buffer, 
					(s->buflen += CHUNK_LEN) + 1);
			n = read(fd, s->buffer + s->buflen - CHUNK_LEN, 
				 CHUNK_LEN);
		} while (n == CHUNK_LEN);
		s->buffer[s->buflen - CHUNK_LEN + n] = 0;
	} else 
		s->buffer[n + s->got] = 0;
	
	if (n == 0)
		eof = 1;

	/* now we shall split the input into lines */
	remaining = split_input(s, eof);

	if (remaining) {
		s->got = strlen(remaining);
		memmove(s->buffer, remaining, s->got + 1);
	} else
		s->got = 0;
		
	if (eof) {
		prtf(0, "\e<error> EOF from server %s : %d", s->name, s->port);
		kill_server(s, 0);	/* do not close, main loop is
					 * supposed to do this, since
					 * we're returing non-zero */
		return -1;
	} else
		return 0;
}

static var *c_status(struct frame *f)
{
	char *s = getsarg(f, "s");

	prt(0, s);
	
	return 0;
}

static var *c_win(struct frame *f)
{
	char *win = getsarg(f, "win");
	char *s = getsarg(f, "s");

	prt(win, s);

	return 0;
}

static var *c_server(struct frame *f)
{
	char *host = getsarg(f, "host");
	int port = getiarg(f, "port");
	char *home = getsarg(f, "home");
	int sock;
	
	sock = make_socket(home);
	
	/* now, after we bound to the right iface, lets look for
	 * our target */
	sock = connect_socket(sock, host, port);

	if (sock != -1) {
		kill_server(&global_server, 1);
		global_server.name = xstrdup(host);
		global_server.port = port;
		global_server.fd = sock;
		sock_link(1, sock, server_handler, &global_server);
		fill_my_addr(sock);
	}

	return new_iv(0, sock);
}

static var *c_shutdown(struct frame *f)
{
	(void)f;
	quit();
	return 0; 	/* bogus */
}

static var *c_detach(struct frame *f)
{
	xsend(CMD_EXIT, "%s", getsarg(f, "reason"));
	return 0;
}

static var *c_load_file(struct frame *f)
{
	load_file(getsarg(f, "name"));
	return 0;
}

static var *c_put_server(struct frame *f)
{
	char *s;
	s = getsarg(f, "s");
	s = xf("%s\r\n", s);
	write(global_server.fd, s, strlen(s));

	return 0;
}

int map_color(char *p, int n, int def);

static var *c_color(struct frame *f)
{
	char *c1 = getsarg(f, "c1"), *c2 = getsarg(f, "c2");

	return new_iv(0,
		      map_color(c1, strlen(c1), LIGHTGRAY) |
		      (map_color(c2, strlen(c2), BLACK) << 4));
}

static var *c_set_status_line(struct frame *f)
{
	int len;
	char *l, *s = getsarg(f, "s");
	
	len = line_count_len(s);
	l = (char *) xmalloc(len + 1);
	line_do_compile(s, l);
	xsend(CMD_SET_STATUS_LINE, "%s", l);
	xfree(l);
	
	return 0;
}

static var *c_kill_window(struct frame *f)
{
	win_kill(getsarg(f, "s"));
	return 0;
}

static var *c_add_window(struct frame *f)
{
	win_add(getsarg(f, "s"));
	return 0;
}

static int irc_tolower(int c)
{
	switch (c) {
	case '[':
		return '{';
	case ']':
		return '}';
	case '\\':
		return '|';
	default:
		if (c >= 'A' && c <= 'Z')
			return c | 0x20;
		else
			return c;
	}
}

static var *c_irc_tolower(struct frame *f)
{
	char *p, *s;
	
	for (s = p = getsarg(f, "s"); *p; p++)
		*p = irc_tolower(*p);
	return new_sv(0, s);
}

int same(const char *n1, const char *n2)
{
	for (; *n1 && *n2; n1++, n2++)
		if (irc_tolower(*n1) != irc_tolower(*n2))
			break;
	return *n1 == *n2;
}

static var *c_same(struct frame *f)
{
	return new_iv(0, same(getsarg(f, "n1"), getsarg(f, "n2")));	
}

static var *c_send_bools(struct frame *f)
{
	(void)f;
	send_bools();
	return 0;
}

static var *c_send_colors(struct frame *f)
{
	(void)f;
	send_colors();
	return 0;
}

void irc_init()
{
	global_server.fd = -1;
	add_cproc("connect_to_server", "string host, int port, string home",
		  c_server);
	add_cproc("status", "string s", c_status);
	add_cproc("puts", "string win, string s", c_win);
	add_cproc("shutdown", "", c_shutdown);
	add_cproc("detach", "string reason", c_detach);
	add_cproc("load_file", "string name", c_load_file);
	add_cproc("color", "string c1, string c2", c_color);
	add_cproc("set_status_line", "string s", c_set_status_line);
	add_cproc("put_server", "string s", c_put_server);
	add_cproc("add_window", "string s", c_add_window);
	add_cproc("kill_window", "string s", c_kill_window);
	add_cproc("same", "string n1, string n2", c_same);
	add_cproc("irc_tolower", "string s", c_irc_tolower);
	add_cproc("send_bools", "", c_send_bools);
	add_cproc("send_colors", "", c_send_colors);
}
