/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: expr.c,v 1.14 2000/08/05 20:19:51 malekith Exp $ */

#include "vm.h"

int test_cond(byte * t, frame * f)
{
	var *v;
	int r = 0;
	
	v = exec(t, f);
	
	switch (v ? v->what : V_VOID) {
	case V_INT:
		r = v->var_int != 0;
		break;
	case V_PTR:
		r = v->var_ptr != 0;
		break;
	case V_STRING:
		r = v->var_string[0] != 0;
		break;
	case V_VOID:
		/* croak(t, "void expression where bool result expected"); */
		r = 0;
		break;
		
	default:
		abort();
	}

	free_var(v);
	
	return r;
}

static char *op1_name(int no)
{
	static char buf[2];

	switch (no) {
	case 'm':
		return "post --";
		break;
	case 'M':
		return "pre --";
		break;
	case 'p':
		return "post ++";
		break;
	case 'P':
		return "pre ++";
		break;
	default:
		buf[0] = no;
		buf[1] = 0;
		return buf;
		break;
	}
}

char *type_name(var *v)
{
	switch (v ? v->what : V_VOID) {
	case V_INT:
		return "int";
	case V_STRING:
		return "string";
	case V_PTR:
		return "ptr";
	case V_VOID:
		return "void";
	default:
		return "unknown";
	}
}

var *do_expr1(vm_token * t, frame * f, int want_lval)
{
	var *x, *r;
	int need_lval = 0, op;

	op = t->u.number;

	need_lval = strchr("mMpP", op) != 0;

	if (want_lval) {
		croak(t, "unary %s won't result in lvalue", op1_name(op));
		return 0;
	}

	if (op == '!')
		return new_iv(0, !test_cond(t->left, f));

	if (need_lval)
		x = exec_lval(t->left, f);
	else
		x = exec(t->left, f);

	if (x == 0 || x->what == V_VOID || x->what == V_STRING ||
	    (x->what == V_PTR && (op == '-' || op == '~'))) {
		croak(t,
		      "unary operator '%s' connot be used with %s values",
		      op1_name(t->u.number), type_name(x));
		if (!need_lval)
			free_var(x);
		return 0;
	}

	if (x->what == V_AD) {
		switch (op) {
		case 'M':
			--(*x->var_string);
			r = new_iv(0, *x->var_string);
			break;
		case 'm':
			r = new_iv(0, *x->var_string);
			(*x->var_string)--;
			break;
		case 'P':
			++(*x->var_string);
			r = new_iv(0, *x->var_string);
			break;
		case 'p':
			r = new_iv(0, *x->var_string);
			(*x->var_string)++;
			break;
		default:
			abort();
		}
		free_var(x); /* V_AD always has to be freed */
		x = 0;
	} else {
		switch (op) {
		case '-':
			r = new_iv(0, -(x->var_int));
			break;
		case '+':
			r = copy(x);
			break;
		case '~':
			r = new_iv(0, ~(x->var_int));
			break;
		case 'M':
			--x->var_int;
			r = copy(x);
			break;
		case 'm':
			r = copy(x);
			--x->var_int;
			break;
		case 'P':
			++x->var_int;
			r = copy(x);
			break;
		case 'p':
			r = copy(x);
			++x->var_int;
			break;
		default:
			abort();
		}
		if (!need_lval)
			free_var(x);
	}
	return r;
}

static int compare_strings(var * x1, var * x2, int op)
{
	int n;
	
	n = strcmp(x1->var_string, x2->var_string);

	switch (op) {
	case '<':
		return n < 0;
	case '>':
		return n > 0;
	case 'M':
		return n >= 0;
	case 'L':
		return n <= 0;
	case 'E':
		return n == 0;
	case 'N':
		return n != 0;
	default:
		abort();
	}
}

static var *do_string2(var *x1, var *x2, vm_token *t, int want_lval)
{
	var *r;

	switch (t->u.number & 0xff) {
	case '+':
		r = new_svn(0, strlen(x1->var_string) + 
				strlen(x2->var_string) + 1);
		strcpy(r->var_string, x1->var_string);
		strcat(r->var_string, x2->var_string);
		break;
	case '*':
		{
			int pl = strlen(x1->var_string), n;
			char *d;
			
			r = new_svn(0, pl * (n = x2->var_int) + 1);
			d = r->var_string;
			while (n--) {
				strcpy(d, x1->var_string);
				d += pl;
			}
			*d = 0;
		}
		break;
	case 0:
		r = copy(x2);
		break;
	default:
		return new_iv(0, compare_strings(x1, x2, t->u.number));
	}

	if (t->u.number & 0x100)
		set_sv(x1, r->var_string);
		
	if (want_lval) {
		free_var(r);
		return x1;
	} else
		return r;
}

char *op2_name(int no)
{
	static char buf[3];

	switch (no) {
	case 'L':
		return "<=";
	case 'E':
		return "==";
	case 'M':
		return ">=";
	case 'N':
		return "!=";
	case 'a':
		return "&&";
	case 'o':
		return "||";
	case 'z'|0x100:
		return "<<=";
	case 'x'|0x100:
		return ">>=";
	case 'z':
		return "<<";
	case 'x':
		return ">>";
	case 0x100:
		return "=";
	default:
		buf[0] = no;
		if (no & 0x100) {
			buf[1] = '=';
			buf[2] = 0;
		} else
			buf[1] = 0;
		return buf;
	}
}

/* a very long but simple function */
static var *do_int2(var *x1, var *x2, vm_token *t, int want_lval)
{
	int res = 0;

	switch (t->u.number & 0xff) {
	case 0:
		res = x2->var_int;
		break;
	case '/':
		if (x2->var_int == 0) {
			croak(t, "division by zero attempted");
			return 0;
		} else
			res = x1->var_int / x2->var_int;
		break;
	case '*':
		res = x1->var_int * x2->var_int;
		break;
	case '+':
		res = x1->var_int + x2->var_int;
		break;
	case '-':
		res = x1->var_int - x2->var_int;
		break;
	case '|':
		res = x1->var_int | x2->var_int;
		break;
	case '&':
		res = x1->var_int & x2->var_int;
		break;
	case '^':
		res = x1->var_int ^ x2->var_int;
		break;
	case 'z':
		res = (unsigned int)x1->var_int << x2->var_int;
		break;
	case 'x':
		res = (unsigned int)x1->var_int >> x2->var_int;
		break;
	case '%':
		if (x2->var_int == 0) {
			croak(t, "modulo by zero attempted");
			return 0;
		} else
			res = x1->var_int % x2->var_int;
		break;
	case 'E':
		res = x1->var_int == x2->var_int;
		break;
	case '<':
		res = x1->var_int < x2->var_int;
		break;
	case '>':
		res = x1->var_int > x2->var_int;
		break;
	case 'N':
		res = x1->var_int != x2->var_int;
		break;
	case 'M':
		res = x1->var_int >= x2->var_int;
		break;
	case 'L':
		res = x1->var_int <= x2->var_int;
		break;
	default:
		abort();
	}


	if (t->u.number & 0x100)
		x1->var_int = res;

	return want_lval ? x1 : new_iv(0, res);
}

var *do_ad(vm_token * t, var * x1, var * x2, int want_lval)
{
	var *tmp;
	int v;

	/* this has to be assigment ! */
	if ((t->u.number & 0x100) == 0)
		abort();
		
	if (x2->what != V_INT) {
		croak(t, "attempted to assign %s to array subscript",
		      type_name(x2));
		free_var(x1);
		free_var(x2);
		return 0;
	}
	tmp = new_iv(0, *x1->var_string);
	do_int2(tmp, x2, t, 1);
	v = *x1->var_string = tmp->var_int;
	free_var(tmp);
	free_var(x2);

	if (want_lval)
		return x1;

	free_var(x1);
	return new_iv(0, v);
}

int undef_hash(vm_token * t, frame * f);

static int check_undef_hash(vm_token * x, frame * f)
{
	vm_token t;
	unpack(&t, x->left);
	
	/* we're only intrested in hashes */
	if (t.id != ID_HASH)
		return 0;
		
	if (undef_hash(&t, f) == 0)
		return 1;
		
	return 0;
}

var *do_expr2(vm_token * t, frame * f, int want_lval)
{
	var *x1, *x2, *r;
	int op;
	char *allow;

	op = t->u.number;

	if (want_lval && (op & 0x100) == 0) {
		croak(t, "binary %s won't result in lvalue", op2_name(op));
		return 0;
	}
	
	if (op & 0x100) {
		x2 = exec(t->right, f);
		x1 = exec_lval(t->left, f);
	} else if (op == 'a') {
		return new_iv(0, test_cond(t->left, f) && 
						test_cond(t->right, f));
	} else if (op == 'o') {
		return new_iv(0, test_cond(t->left, f) || 
						test_cond(t->right, f));
	} else if (op == '@') {
		x1 = exec(t->left, f);
		op = regexp_do(x1->var_string, t->right, f);
		free_var(x1);
		return new_iv(0, op);
	} else {
		x1 = exec(t->left, f);
		x2 = exec(t->right, f);
	}
	
	if (op == 0x100 && (x2 == 0 || x2->what == V_VOID) 
	    && check_undef_hash(t, f)) {
		free_var(x2);
		return 0;
	}
			    
	if ((x1 == 0 || x2 == 0) ||
	    (x1->what == V_VOID) || (x2->what == V_VOID)) {
		croak(t, "void expressions cannot be used in binary '%s'",
			op2_name(op));
		if ((op & 0x100) == 0)
			free_var(x1);
		free_var(x2);
		return 0;
	}

	if (x1->what == V_AD)
		return do_ad(t, x1, x2, want_lval);
	
	if (x1->what == V_INT && x2->what == V_PTR)
		allow = op & 0x100 ? 0 : "+";
	else if (x1->what == V_PTR && x2->what == V_INT)
		allow = "-+";
	else if (x1->what == V_PTR && x2->what == V_PTR)
		allow = op & 0x100 ? 0 : "-<>ENLM";
	else if (x1->what == V_STRING && x2->what == V_INT)
		allow = "*";
	else if (x1->what == V_STRING || x2->what == V_STRING)
		if (x1->what != x2->what)
			allow = 0;
		else 
			allow = "+<>ENLM";
	else if (x1->what == V_INT && x2->what == V_INT)
		goto two_ints;
	else
		abort();
	
	if ((op == 0x100 && x1->what != x2->what) ||
	    (allow == 0 || strchr(allow, op & 0xff) == 0)) {
		croak(t, "cannot do '%s' %s '%s'", type_name(x1),
			op2_name(op), type_name(x2));
		return 0;
	}
			
two_ints:
	if (op & 0x100 && strchr("/*-+%zx|&^",op&0xff)==0)
		abort(); /* no such thing shall exist */
		
	if (x1->what == V_STRING || x2->what == V_STRING)
		r = do_string2(x1, x2, t, want_lval);
	else
		r = do_int2(x1, x2, t, want_lval);

	if (x1->what == V_PTR || x2->what == V_PTR)
		r->what = V_PTR;
		
	if ((op & 0x100) == 0)
		free_var(x1);
	free_var(x2);

	return r;
}

