/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: parser.h,v 1.3 2000/08/04 21:28:14 malekith Exp $ */

#ifndef _SCHIZA_PARSER_H
#define _SCHIZA_PARSER_H 1

#include "common.h"

#define CTX struct parser_state

struct frame;

typedef struct token {
	struct token *next, *left, *right;
	/* next is for normal flow control, top, left & right are for
	 * if's and such */
	struct {
	 	char *name;	/* strduped name/caption if needed */
		int number;	/* int caption for int constas */
		struct token *top;
	} u;
	
	/* mainly for REs */
	void *data_ptr;
	int data_len; 
	
	/* error info */
	unsigned int line;
	unsigned char file_idx;
	
	/* what is it */
	unsigned char id; 
} token;

struct parser_state {
	char *p;
	FILE *f;
	char line[4096];
	const char *fn;
	int file_idx;
	int line_no;
	struct frame *frm;
	
	int type;
	int iv;
	char *sv;
	int leave_it;

	int hash;
	char *re_char;
	int stop1, stop2, stop3;
};

/* CTX -> type */
enum {
T_INT = 1,
T_STRING,
T_WORD,
T_OP,
};

/* parser.c */
void poops(CTX *ctx,  const char *fmt, ...);
token *make_token(CTX * ctx, int id, char *name, int numb);
token *get_instruction(CTX * ctx);
token *link_end(token ** h, token ** x, token * what);

/* pexpr.c */
token *get_expression(CTX * ctx);

/* token.c */
int get_token(CTX * ctx);
char *get_regexp(CTX *ctx);

/* keyword.c */
int is_keyword(char *s);
token *handle_keyword(CTX *ctx, char *word);

/* bytecode.c */
byte *pack(token *t, int more, byte **more_ptr, int *total_size);
void free_bc(byte *code);

void kill_token(token * t);
const char *regexp_make(const char *string, token *t);

#endif /* _SCHIZA_PARSER_H */
