/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: vm.h,v 1.2 2000/08/04 21:28:14 malekith Exp $ */

#ifndef _SCHIZA_VM_H
#define _SCHIZA_VM_H 1

#include "common.h"
#include "var.h"

/*
 * -----------------------------------------------------------------------
 * token
 * -----------------------------------------------------------------------
 */
typedef struct vm_token {
	byte *next, *left, *right, *top;
	/* next is for normal flow control, top, left & right are for
	 * if's and such */
	struct {
	 	char *name;	/* strduped name/caption if needed */
		int number;	/* int caption for int constas */
		byte *top;
	} u;
	void *data_ptr;
	u32 hash;
	
	/* error info */
	unsigned int line;
	unsigned char file_idx;
	
	/* what is it */
	unsigned char id; 
} vm_token;


/*
 * -----------------------------------------------------------------------
 * frame
 * -----------------------------------------------------------------------
 */
struct regexp_info;
typedef struct frame {
	struct hash *stack;	/* one hash with all vars */
	struct hash *hashes;	/* linked list */
	
	int break_lev, continue_lev, return_lev;
	
	struct frame *parent;
	struct frame *global;
	struct hash *procs;
	
	struct regexp_info *regexp;
	struct vm_token *active_token;	/* set only in call to cprocs,
					 * so they can do croak() */
} frame;



/*
 * -----------------------------------------------------------------------
 * protos
 * -----------------------------------------------------------------------
 */
 
/* hash_vm.c */
/* kill all hashes in frame */
void kill_frame(frame * f);
/* for users hashes */
void hash_add_one(frame * f, const char *name, int type);
var *hash_entry(vm_token *t, frame *f);
var *hash_walk(vm_token *t, frame *f);

/* script.c */
void croak(vm_token *t, const char *fmt, ...);
char *input_file_name(vm_token *t);

/* expr.c */
int test_cond(byte * t, frame * f);
var *do_expr2(vm_token * t, frame * f, int lval);
var *do_expr1(vm_token * t, frame * f, int lval);
char *type_name(var*);

/* flow.c */
void setup_frame(frame *ret, frame *parent);
var *exec(byte *t, frame *f);
var *exec_nx(byte *t, frame *f, byte **next);
void compile_file(const char *fn);
var *exec_lval(byte *t, frame *f);
var *exec_loop(byte *t, frame *f);

/* magic.c */
var *get_magic_var(vm_token *t, frame *f);

/* regexp.c */
var *regexp_get(vm_token *t, frame *f, int no);
int regexp_do(const char * x1, byte *x2, frame * f);
void regexp_info_free(struct regexp_info *r);

/* bytecode.c */
void unpack(vm_token *t, byte *src);
void free_bc(byte *code);

/* var.c */
	/* doesn't copy, p becomes lval */
void drop_on_stack(var *p, frame *f);	
	/* look in f->stack, global.stack */
var *get_var(const char *name, frame *f);
var *get_var_h(const char *name, frame *f, u32 h);

#endif /* _SCHIZA_VM_H */
