/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: hash.c,v 1.14 2000/08/06 18:41:15 malekith Exp $ */

#include "common.h"
#include "var.h"
#include "proc.h"

hash *hash_new(int type)
{
	hash *h;

	h = (hash *) xmalloc(sizeof(*h));
	memset(h, 0, sizeof(*h));
	h->entries = (var **) xmalloc((h->size = 16) * sizeof(var *));
	memset(h->entries, 0, 16 * sizeof(var *));
	h->num_entries = 0;
	h->type = type;

	h->free_entry = free_var;
	
	return h;
}

static void check_denisty(hash * h)
{
	int newsize, i;
	var **e, *p, *tmp;
	
	if (h->num_entries / h->size < 20)
		return;
		
	newsize = h->num_entries / 4;
	e = (var **) xmalloc(newsize * sizeof(var *));
	memset(e, 0, newsize * sizeof(var *));
	
	for (i = 0; i < h->size; i++)
		for (p = h->entries[i]; p; p = tmp) {
			u32 x;
			
			x = xhash(p->name) % newsize;
			tmp = p->next;
			p->next = e[x];
			e[x] = p;
		}
		
	xfree(h->entries);
	h->size = newsize;
	h->entries = e;
}

u32 xhash(const unsigned char *key)
{
	u32 x = 0;
	
	while (*key)
		x = x * 33 + *key++;
	return x;
}

var *hash_find_ex2(hash *h, const unsigned char *key, u32 x)
{
	var *p;
	
	for (p = h->entries[x % h->size]; p; p = p->next)
		if (strcmp(p->name, key) == 0)
			break;
	return p;
}

var *hash_find_ex(hash *h, const unsigned char *key)
{
	return hash_find_ex2(h, key, xhash(key));
}


void hash_add(hash *h, var *p)
{
	int x = xhash(p->name) % h->size;
	
	p->next = h->entries[x];
	h->entries[x] = p;
	h->num_entries++;
	check_denisty(h);
}

void hash_kill_one(hash * h)
{
	var *p, *tmp;
	int i;
	
	for (i = 0; i < h->size; i++)
		for (p = h->entries[i]; p; p = tmp) {
			tmp = p->next;
			h->free_entry(p);
		}
		
	xfree(h->entries);
	xfree(h->name);
	xfree(h);
}

int hash_del(hash * h, const char *key)
{
	var *o, *p;
	u32 i;

	o = hash_find_ex(h, key);
	if (o == 0)
		return -1;
		
	i = xhash(key) % h->size;
	p = h->entries[i];
	
	if (p == o)
		h->entries[i] = p->next;
	else {
		while (p && p->next) {
			if (p->next == o)
				break;
			p = p->next;
		}
		if (p->next == 0)
			abort();
		p->next = o->next;
	}
	
	h->free_entry(o);
		
	h->num_entries--;

	return 0;
}
