/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: token.c,v 1.11 2000/08/04 21:28:14 malekith Exp $ */

#include "parser.h"

static int more_input(CTX *ctx)
{
	char *p;

	p = ctx->p;
	again:
	while (isspace(*p))
		p++;
	if (*p == 0) {
		if (ctx->f == 0 || 
		    fgets(ctx->line, sizeof(ctx->line), ctx->f) == 0) {
			ctx->p = p;
			return -1;
		}
		p = ctx->p = ctx->line;
		if (p[strlen(p) - 1] == '\n')
			p[strlen(p) - 1] = 0;
		ctx->line_no++;
		goto again;
	}
	if (*p == '#') {
		*p = 0;
		goto again;
	}
	ctx->p = p;
	return 0;
}

static char *parse_word(CTX * ctx)
{
	char *p, *r;

	p = ctx->p;
	while (((*p | 0x20) >= 'a' && (*p | 0x20) <= 'z')
	       || (*p >= '0' && *p <= '9') || *p == '_' || *p == '$')
		p++;
		
	r = xstrndup(ctx->p, p - ctx->p);
	
	ctx->p = p;
	
	return r;
}

static int get_one_char(CTX *ctx)
{
	int r;
	
	if (*ctx->p == '\\') {
		ctx->p++;
		switch (*ctx->p++) {
		case '\\':
			return '\\';
		case 't':
			return '\t';
		case 'n':
			return '\n';
		case 'r':
			return '\r';
		case 'f':
			return '\f';
		case 'b':
			return '\b';
		case 'a':
			return '\a';
		case 'e':
			return '\e';
		case 'c':
			return *ctx->p++ & 31;
		case '"':
			return '"';
		case 'x':
			r = strtoul(ctx->p, &ctx->p, 16);
			if (r & 0xffffff00)
				poops(ctx, "\\x value out of range (%x)", r);
			return r&0xff;
		case '0' ... '7':
			r = strtoul(ctx->p, &ctx->p, 8);
			if (r & 0xffffff00)
				poops(ctx, "\\0 value out of range (%x)", r);
			return r&0xff;
		case '$':
			ctx->p--;
			return '\\';
		default:
			return '?';
		}
	} else
		return *ctx->p++;
}

static int re_get_one_char(CTX * ctx)
{
	return *ctx->p++;
}

static char *parse_string(CTX * ctx, int stop)
{
	int len;
	char *p, *r;

	p = ctx->p;
	for (len = 0; *ctx->p && *ctx->p != stop; len++)
		get_one_char(ctx);

	if (*ctx->p == 0)
		return 0;

	ctx->p = p;
	p = r = (char *) xmalloc(len + 1);
	while (*ctx->p && *ctx->p != stop)
		*p++ = get_one_char(ctx);
	*p = 0;

	ctx->p++;
	if (more_input(ctx) == 0 && *ctx->p == stop) {
		char *r2;

		ctx->p++;
		r2 = parse_string(ctx, stop);
		p = (char *) xmalloc(strlen(r) + strlen(r2) + 1);
		strcpy(p, r);
		strcat(p, r2);
		xfree(r);
		xfree(r2);
		r = p;
	}
	
	return r;
}

char *get_regexp(CTX *ctx)
{
	int len, stop;
	char *p, *r;

	if (more_input(ctx))
		return 0;
		
	stop = *ctx->p++;
	
	p = ctx->p;
	for (len = 1; *ctx->p && *ctx->p != stop; len++)
		re_get_one_char(ctx);

	if (*ctx->p == 0) {
		poops(ctx, "not finished regexp");
		return 0;
	}

	ctx->p = p;
	p = r = (char *) xmalloc(len + 1);
	*p++ = 0x01;
	while (*ctx->p && *ctx->p != stop)
		*p++ = re_get_one_char(ctx);
	*p = 0;

	ctx->p++;

	while (more_input(ctx) == 0) 
		switch (*ctx->p) {
		case 'i':
			ctx->p++;
			r[0] |= 0x02;
			break;
		case 'm':
			ctx->p++;
			r[0] |= 0x04;
			break;
		case 'x':
			ctx->p++;
			r[0] |= 0x08;
		default:
			goto out;
		}
	
      out:
	if (more_input(ctx) == 0 && *ctx->p == stop) {
		char *r2;

		r2 = get_regexp(ctx);
		p = (char *) xmalloc(strlen(r+1) + strlen(r2+1) + 2);
		strcpy(p, r);
		strcat(p, r2+1);
		p[0] |= r2[0];
		xfree(r);
		xfree(r2);
		r = p;
	}
	return r;
}

static int do_stops(CTX *ctx)
{
	if (ctx->stop1) {
		ctx->type = T_OP;
		ctx->iv = '(';
		ctx->stop2 = ctx->stop1;
		ctx->stop1 = 0;
		return 1;
	}
	if (ctx->stop2) {
		ctx->type = T_STRING;
		ctx->sv = parse_string(ctx, ctx->stop2);
		ctx->stop3 = ctx->stop2;
		ctx->stop2 = 0;
		return 1;
	}
	if (ctx->stop3) {
		ctx->type = T_OP;
		ctx->iv = ')';
		ctx->stop3 = 0;
		return 1;
	}
	return 0;
}

static void choose_token(CTX *ctx, int hash)
{
	char *p;
	
	p = ctx->p++;

	switch (*p) {
	case '"':
		ctx->type = T_STRING;
		ctx->sv = parse_string(ctx, '"');
		break;
	case '`':
		ctx->type = T_WORD;
		ctx->sv = xstrdup("__back_tick");
		ctx->stop1 = '`';
		break;
	case '?':
		ctx->type = T_WORD;
		ctx->sv = xstrdup("__question_mark");
		ctx->stop1 = '?';
		break;
	case ':':
		ctx->type = T_WORD;
		ctx->sv = xstrdup("__colon");
		ctx->stop1 = ':';
		break;
	case '.':
		ctx->type = T_WORD;
		ctx->sv = xstrdup("__dot");
		ctx->stop1 = '.';
		break;
	case '0':
		p++;
		ctx->type = T_INT;
		if (*p == 'x' || *p == 'X') {
			ctx->p = ++p;
			ctx->iv = strtoul(ctx->p, &ctx->p, 16);
		} else if (*p >= '0' && *p <= '7') {
			ctx->iv = strtoul(ctx->p, &ctx->p, 8);
		} else if (isalnum(*p) || *p == '_') {
			poops(ctx, "shit after zero");
			ctx->type = 0;
		} else {
			ctx->iv = 0;
		}
		break;
	case '1' ... '9':
		ctx->p--;
		ctx->type = T_INT;
		ctx->iv = strtoul(ctx->p, &ctx->p, 10);
		break;
	case '\'':
		ctx->type = T_INT;
		ctx->iv = get_one_char(ctx);
		if (*ctx->p == '\'')
			ctx->p++;
		break;	
	case 'a' ... 'z':
	case 'A' ... 'Z':
	case '_':
	case '$':	/* $ is normal char in identifiers */
		ctx->p--;
		ctx->type = T_WORD;
		ctx->sv = parse_word(ctx);
		if (!is_keyword(ctx->sv) && hash != 10)
			ctx->hash = 1;
		else if (strcmp(ctx->sv, "walk") == 0)
			ctx->hash = 10;
		break;	
	case '!':
		ctx->type = T_OP;
		if (*++p == '=') {
			ctx->p++;
			ctx->iv = 'N';
		} else
			ctx->iv = '!';
		break;	
	case '<':
	case '>':
		ctx->type = T_OP;
		if (p[1] == '=') {
			ctx->p++;
			ctx->iv = *p == '<' ? 'L' : 'M';
		} else if (p[0] == p[1]) {
			ctx->p++;
			ctx->iv = *p == '<' ? 'z' : 'x';
			if (p[2] == '=') {
				ctx->p++;
				ctx->iv |= 0x100;
			}
		} else {
			ctx->iv = *p;
		}
		break;
	case '{':
		if (hash == 1) {
			ctx->type = T_OP;
			ctx->iv = '[';
			ctx->hash = 2;
			break;
		}
	case '}':
		if (hash == 3) {
			ctx->type = T_OP;
			ctx->iv = ']';
			break;
		}
	case '(':
	case ')':
	case '[':
	case ']':
	case ',':
	case ';':
	case '~':
	case '@':	/* regexp */
		ctx->type = T_OP;
		ctx->iv = *p;
		break;	
	case '=':
		ctx->type = T_OP;
		if (p[0] == p[1]) {
			ctx->p++;
			ctx->iv = 'E';
		} else
			ctx->iv = '=';
		break;
	case '-':
	case '+':
	case '|':
	case '&':
		if (p[0] == p[1]) {
			ctx->p++;
			ctx->type = T_OP;
			ctx->iv = 
				*p == '&' ? 'a' 
				: *p == '|' ? 'o'
				: *p == '-' ? 'm'
				: 'p';
			break;
		}
	case '%':
	case '*':
	case '/':
	case '^':
		ctx->type = T_OP;
		ctx->iv = *p;
		if (*++p == '=') {
			ctx->p++;
			ctx->iv |= 0x100;
		}
		break;
	default:
		poops(ctx, "illegal character '%c'", *p);
		break;
	}
}

int get_token(CTX * ctx)
{
	int hash;

	if (ctx->leave_it) {
		ctx->leave_it = 0;
		return ctx->type != 0;
	}

	if (do_stops(ctx))
		return 1;
		
	hash = ctx->hash;
	
	switch (hash) {
	case 1: /* after word (`{' shall be interpreted as a `[') */
	case 3: /* after returning word (return `}' as `]') */
	case 10: /* walk occured */
		ctx->hash = 0;
		break;
	case 2: /* after interpreting `{' as `[' (return word as T_STRING) */
		break; /* handled below */
	case 0: /* normal */
		break; 
	default:
		abort();
	}
	 
	if (ctx->sv)
		xfree(ctx->sv);
	ctx->type = 0;
	ctx->sv = 0;
	ctx->iv = 0;

	if (more_input(ctx))
		return 0;
	
	if (hash == 2) {
		ctx->type = T_STRING;
		ctx->sv = parse_word(ctx);
		ctx->hash = 3;
		return 1;
	}
	
	choose_token(ctx, hash);
	
	if (ctx->type == T_OP && ctx->iv == 0)
		abort();
	
	return 1;
}
