/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: sniffy.h,v 1.2 2000/08/27 18:32:51 malekith Exp $ */

#include "logic.h"

extern "C" {
#include "lib.h"
#include "proto.h"
};

// curses.cc
extern int s_width, s_height;
extern int scroll_top, scroll_bottom;

enum {
	key_special = 0x10000,
	key_left, key_right, key_up, key_down, key_home,
	key_end, key_ins, key_del, key_pgdn, key_pgup,
	key_5,
	key_f0 = key_special + 0x100,
	key_meta = 0x100000
};

#define key_f(a) (key_f0 + (a))
#define key_ctrl(a) (a & 0x9f)

void init_curses();
void kill_curses();

void set_color(int vga);
int print(const char *str, ...) __attribute__ ((format(printf, 1, 2)));
void print(int ch, int color = -1);
void go(int x, int y);

void f_repaint();
void repaint();
int get_frame(int x, int y);
void scroll_up(int n);
void draw_frame(int c, int x, int y, int w, int h);

int _keyb();
#ifndef getch
inline int getch() { return _keyb(); }
#endif

class InputLine {
public:
	void draw();
	void handle_input();
	void handle_key(int);
	void self_insert(int);
	void call_perl(const char*);
	void fix_pos();
	const char *prompt() { return prompt_string; }
	char caption[2000];
	char prompt_string[100];
	int curs_pos, scr_offset;
	InputLine() : curs_pos(0), scr_offset(0) 
		{ caption[0] = 0; strcpy(prompt_string, "% "); }
};

class IrcWindow {
	void draw_line(int i);
	void draw_lines(int n);
public:
	IrcWindow() : next(0), offset(-1), changed(0), no(0) { }
	IrcWindow *next;
	IrcBuffer buffer;
	InputLine iline;
	int offset;
	int scroll_jump();
	void add(const char *str);
	void draw();
	void pgup();
	void pgdn();
	void clr() { buffer.clr(); draw(); }
	
	char channel[100];
	bool changed;
	int no;
};

void message_box(const char *str_o, const char *title);

/* window.cc */
void kill_window(const char *ch);
void clear_win(const char *win);
void bring_to_top(const char *ch);
void add_window(const char *chan);
void select_win(IrcWindow *w);
void send_out(const char *ch, const char *msg);
void put_line(int x, int y, const char *p);
IrcWindow *find_win(const char *ch);
extern IrcWindow *win_head, *win_topmost, *win_status;
// for alt+1 [1], alt+2 ... alt+shift+0 [10], alt+shift+1 [11] 
extern IrcWindow *win20[20]; 

/* inputline.cc */
int menu_handle_key(int);
int menu_active();
extern InputLine *input_line;
